/****************************************************************************
 *
 *     AGE -- Alignment with Gap Excision
 *     Copyright (C)  Alexej Abyzov
 *                
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the Creative Commons license
 * (Attribution-NonCommerical).
 * See term at http://creativecommons.org/licenses/by-nc/2.5/legalcode
 *
 * Author: Alexej Abyzov
 */


#define AGE_TIME


//--- C/C++ includes ---
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#ifdef AGE_TIME
#include <sys/time.h>
#endif
using namespace std;

const static int GAP_OPEN   =  -2;
const static int GAP_EXTEND =  -1;
const static int MATCH      =   1;
const static int MISMATCH   =  -1;

const static unsigned char DIAGONAL   =   1;
const static unsigned char HORIZONTAL =   2;
const static unsigned char VERTICAL   =   3;


// Nucleotide sequence
// Allowed nucleotides are A,C,T,G,U,X,N in both lower and upper cases
// Other nucleotides (R,Y,M,K,S,W,B,D,H,V) are also alowed by not scored
class Sequence
{
private:
  string _seq,_name;
  int    _start;     // 1 based coordinate
  bool   _reverse;

public:
  inline string &name()      { return _name; }
  inline string &sequence()  { return _seq; }
  inline int     start()     { return _start; }  // 1 based coordinate
  inline int     reverse()   { return _reverse; }

public:
  Sequence(const char *seq) : _seq(seq), _name(""),_start(1),_reverse(false)
  {
    string ret = "";
    for (int i = 0;i < _seq.length();i++) {
      char c = _seq[i];
      if (c == ' ') continue;      // Space
      else if (isNuc(c)) ret += c; // Nucleotide
      else if (!isGap(c)) {        // Gap
	cerr<<"Unrecognized nucleotide '"<<c<<"' in sequence."<<endl;
	_seq = "";
	return;
      }
    }
    _seq = ret;
  }
  
  Sequence(string file,int start = 1,int end = -1) : _seq(""),  _name(""),
						     _start(1), _reverse(false)
  {
    if (start > 0 && end > 0 && end < start) {
      cerr<<"Invalid coordinates ("<<start<<", "<<end<<"). "
	  <<"Start is larger than end."<<endl;
      cerr<<"No sequence read!"<<endl;
      return;
    }

    if (start >= 0) _start = start;

    ifstream in(file.c_str());
    if (!in) {
      cerr<<"Can't open input file '"<<file<<"'."<<endl;
      return;
    }

    char l[1024];
    while (!in.eof() && in.getline(l,1024)) 
      if (in.gcount() != 0) break;

    if (l[0] != '>') {
      cerr<<"Can't find fasta header in the file '"<<file<<"'."<<endl;
      return;
    }
    _name = &l[1];

    string line = "";
    int n_passed = 0;
    while (!in.eof()) {
      in>>line;
      string checked = "";
      for (int i = 0;i < line.length();i++) {
	char c = line[i];
	if (isNuc(c)) checked += c;
	else if (!isGap(c)) {
	  cerr<<"Unrecognized nucleotide '"<<c
	      <<"' in file '"<<file<<"'."<<endl;
	  _seq = "";
	  return;
	}
      }
      int n_checked = checked.length();
      if (n_checked == 0) break;

      int new_passed = n_passed + n_checked;
      if (new_passed < start) { ; // Not reached start yet
      } else if (n_passed < start && new_passed >= end) { // Spans whole region
	_seq += checked.substr(start - n_passed - 1,end - start + 1);
      } else if (n_passed < start && new_passed >= start) { // Spans start only
	_seq += checked.substr(start - n_passed - 1);
      } else if (end < start) { // No end
	_seq += checked;
      } else if (new_passed < end) { // Not reached end yet
	_seq += checked;
      } else if (n_passed < end && new_passed >= end) { // Spans end only
	_seq += checked.substr(0,end - n_passed);
      }
      n_passed = new_passed;
      line = "";
    }
    in.close();
    
    //cout<<_seq.length()<<" nucs."<<endl;
  }
private:
  Sequence(string &seq,string &name,int start,bool reverse) : _seq(seq),
							      _name(name),
							      _start(start),
							      _reverse(reverse)
  {}

public:
  Sequence *clone()
  {
    return new Sequence(_seq,_name,_start,_reverse);
  }

  static char complement(char c)
  {
    if      (c == 'a') return 't';
    else if (c == 'c') return 'g';
    else if (c == 't') return 'a';
    else if (c == 'g') return 'c';
    else if (c == 'A') return 'T';
    else if (c == 'C') return 'G';
    else if (c == 'T') return 'A';
    else if (c == 'G') return 'C';
    else if (c == 'u') return 'a';
    else if (c == 'U') return 'A';
    return c;
  }

  static bool isNuc(char c)
  {
    if (c == 'A' || c == 'a' ||
	c == 'T' || c == 't' ||
	c == 'C' || c == 'c' ||
	c == 'G' || c == 'g' ||
	c == 'U' || c == 'u' ||
	c == 'X' || c == 'x' ||
	c == 'N' || c == 'n' ||
	c == 'R' || c == 'r' ||
	c == 'Y' || c == 'y' ||
	c == 'M' || c == 'm' ||
	c == 'K' || c == 'k' ||
	c == 'S' || c == 's' ||
	c == 'W' || c == 'w' ||
	c == 'B' || c == 'b' ||
	c == 'D' || c == 'd' ||
	c == 'H' || c == 'h' ||
	c == 'V' || c == 'v') return true;
    return false;
  }

  static char gap() { return '-'; }

  static bool isGap(char c)
  {
    if (c == '-') return true;
    return false;
  }

  static int matchNucs(char a,char b)
  {
    char aa = toupper(a), bb = toupper(b);
    if ((aa == 'A' || aa == 'C' || aa == 'T' || aa == 'G' || aa == 'U') &&
	(bb == 'A' || bb == 'C' || bb == 'T' || bb == 'G' || bb == 'U')) 
      if (aa == bb)                    return MATCH;
      else if (aa == 'U' && bb == 'T') return MATCH;
      else if (bb == 'U' && aa == 'T') return MATCH;
      else                             return MISMATCH;
    return 0;
  }

  static bool sameNuc(char a,char b)
  {
    char aa = toupper(a), bb = toupper(b);
    if ((aa == 'A' || aa == 'C' || aa == 'T' || aa == 'G' || aa == 'U') &&
	(bb == 'A' || bb == 'C' || bb == 'T' || bb == 'G' || bb == 'U') &&
	aa == bb) return true;
    return false;
  }

  void revcom()
  {
    _start += _seq.length() - 1;
    _reverse = !_reverse;
    string ret = "";
    for (int i = _seq.length() - 1;i >= 0;i--) ret += complement(_seq.at(i));
    _seq = ret;
  }

  Sequence *substr(int start,int end) // 1 based coordinates
  {
    if (start > 0 && end > 0 && start > end) return NULL;
    if (start <= 0)                        start = 1;
    if (end   <= 0 || end > _seq.length()) end   = _seq.length();
    int s = start, e = end;
    if (_reverse) {
      s = _seq.length() - e + 1;
      e = _seq.length() - s + 1;
    }
    string tmp = "";
    if (_seq.length() > 0) tmp = _seq.substr(s - 1,e - s + 1);
    return new Sequence(tmp,_name,_start + start - 1,_reverse);
  }
};

class Scorer
{
private:
  short _scores[256][256];

public:
  Scorer(short match,short mismatch)
  {
    memset(_scores,0,256*256*sizeof(short));
    for (int i1 = 0;i1 < 256;i1++) {
      char c1 = (char)i1;
      if (Sequence::complement(c1) == c1) continue;
      for (int i2 = 0;i2 < 256;i2++) {
	char c2 = (char)i2;
	if (Sequence::complement(c2) == c2) continue;
	if (Sequence::sameNuc(c1,c2)) _scores[c1][c2] = match;
	else                          _scores[c1][c2] = mismatch;
      }
    }
  }
  
  inline short getScore(char c1,char c2) { return _scores[c1][c2]; }
};


class Block
{
private:
  Block *_next, *_prev;
  int    _start1,_start2,_length;
  string _desc;

public:
  Block(int st1,int st2,int l,string d = "") : _next(NULL), _prev(NULL),
					       _start1(st1), _start2(st2),
					       _length(l), _desc(d)
  {}

  string toString()
  {
    string ret = _desc;
    if (ret.length() > 0) ret += " ";
    ret += "block\t";
    stringstream tmp;
    tmp.str(""); tmp<<_start1;
    ret += tmp.str();  ret += "\t";
    tmp.str(""); tmp<<_start2;
    ret += tmp.str();  ret += "\t";
    tmp.str(""); tmp<<_length;
    ret += tmp.str();
    return ret;
  }

  Block *next(Block *b)
  {
    if (_next) return NULL;
    _next = b;
    b->prev(this);
    return _next;
  }

  Block *prev(Block *b)
  {
    if (_prev) return NULL;
    _prev = b;
    b->next(this);
    return _prev;
  }

  inline Block *next()   { return _next; }
  inline Block *prev()   { return _prev; }
  inline int    start1() { return _start1; }
  inline int    start2() { return _start2; }
  inline int    length() { return _length; }
};

class AGEaligner
{
private:
  Sequence       &_s1,    &_s2;
  string         &_seq1,  &_seq2;
  string         _ali1,   _ali2;
  int            _len1,   _len2;
  int            score_n1,score_n2;
  long           score_size;
  unsigned short *f_score,*r_score;
  unsigned char  *trace;

  int _lg1,_lg2,_rg1,_rg2; // Coordiantes of the big gap from left side
  int _lg1_alt,_lg2_alt,_rg1_alt,_rg2_alt; // -||- from right side
  int _max; // Maximum score

public:
  AGEaligner(Sequence &s1,Sequence &s2) :
    _s1(s1), _s2(s2),
    _seq1(_s1.sequence()), _len1(_seq1.length()), _ali1(""),
    _seq2(_s2.sequence()), _len2(_seq2.length()), _ali2(""),
    _lg1(-1),     _lg2(-1),     _rg1(-1),     _rg2(-1),
    _lg1_alt(-1), _lg2_alt(-1), _rg1_alt(-1), _rg2_alt(-1),
    _max(0)
  {
    score_n1 = _len1 + 2;
    score_n2 = _len2 + 2;
    score_size = (long)score_n1*(long)score_n2;

    // Scoring matrices
    f_score = new unsigned short[score_size];
    r_score = new unsigned short[score_size];
    memset(f_score,0,score_size*sizeof(unsigned short));
    memset(r_score,0,score_size*sizeof(unsigned short));

    // Trace matrix
    trace   = new unsigned char[score_size];
    memset(trace,0,score_size*sizeof(unsigned char));
  }

  ~AGEaligner()
  {
    delete[] f_score;
    delete[] r_score;
    delete[] trace;
  }

  bool align()
  {
    if (_len1 <= 0 || _len2 <= 0) return false;

#ifdef AGE_TIME
    timeval score_s,score_e;
    timeval max_s,  max_e;
    timeval find_s, find_e;
    timeval ali_s,  ali_e;

    gettimeofday(&score_s,NULL);
#endif

    calcScores();

#ifdef AGE_TIME
    gettimeofday(&score_e,NULL);
    gettimeofday(&max_s,NULL);
#endif

    calcMaxima();

#ifdef AGE_TIME
    gettimeofday(&max_e,NULL);
    gettimeofday(&find_s,NULL);
#endif

    _max = findUnaligned(_lg1, _lg2,    _rg1,    _rg2);
    findUnaligned(_lg1_alt,_lg2_alt,_rg1_alt,_rg2_alt,false);

#ifdef AGE_TIME
    gettimeofday(&find_e,NULL);
    gettimeofday(&ali_s,NULL);
#endif

    Block *bls = findBlocks(_lg1,_lg2,_rg1,_rg2);
    findAlignment(bls);

    // Deleting blocks
    while (bls) {
      Block *tmp = bls;
      bls = bls->next();
      delete tmp;
    }

#ifdef AGE_TIME
    gettimeofday(&ali_e,NULL);

    // cout<<"Scoring  "<<score_e.tv_sec - score_s.tv_sec +
    //   (score_e.tv_usec - score_s.tv_usec)/1000000.<<" s"<<endl;
    // cout<<"Maxima   "<<max_e.tv_sec - max_s.tv_sec +
    //   (max_e.tv_usec - max_s.tv_usec)/1000000.<<" s"<<endl;
    // cout<<"Finding  "<<find_e.tv_sec - find_s.tv_sec +
    //   (find_e.tv_usec - find_s.tv_usec)/1000000.<<" s"<<endl;
    // cout<<"Aligning "<<ali_e.tv_sec - ali_s.tv_sec +
    //   (ali_e.tv_usec - ali_s.tv_usec)/1000000.<<" s"<<endl;
#endif

    return true;
  }

  void printAlignment()
  {
    if (_ali1.length() != _ali2.length()) {
      cerr<<"Error in printing alignment!"<<endl;
      return;
    }

    const static string UNALIGNED_MESSSAGE = "UNALIGNED REGION";

    cout<<endl;
    cout<<"First  seq => "
	<<setw(9)<<_seq1.length()<<" nucs '"<<_s1.name()<<"'"<<endl;
    cout<<"Second seq => "
	<<setw(9)<<_seq2.length()<<" nucs '"<<_s2.name()<<"'"<<endl<<endl;

    int ind1 = -1,ind2  = -1,ind3  = -1,ind4  = -1;
    int first_nuc1  = -1,first_nuc2  = -1,first_nuc3  = -1,first_nuc4  = -1;
    int second_nuc1 = -1,second_nuc2 = -1,second_nuc3 = -1,second_nuc4 = -1;
    int n_nuc1 = 0, n_nuc2 = 0;
    for (int i = 0;i < _ali1.length();i++) {
      bool isNuc1 = Sequence::isNuc(_ali1[i]);
      bool isNuc2 = Sequence::isNuc(_ali2[i]);
      if (isNuc1) n_nuc1++;
      if (isNuc2) n_nuc2++;
      if (isNuc1 && isNuc2) {
	first_nuc4  = n_nuc1;
	second_nuc4 = n_nuc2;
	ind4        = i;
	if (first_nuc1 < 0) {
	  first_nuc1  = n_nuc1;
	  second_nuc1 = n_nuc2;
	  ind1        = i;
	}
      }
      if (n_nuc1 == _lg1 && n_nuc2 == _lg2) {
	first_nuc2  = n_nuc1;
	second_nuc2 = n_nuc2;
	ind2        = i;
      }
      if (n_nuc1 == _rg1 && n_nuc2 == _rg2) {
	first_nuc3  = n_nuc1;
	second_nuc3 = n_nuc2;
	ind3        = i;
      }
    }

    int n_ali = 0,n_id = 0,n_gap = 0;
    if (ind2 >= 0) n_ali += ind2 - ind1 + 1;
    if (ind3 >= 0) n_ali += ind4 - ind3 + 1;
    for (int i = ind1;i <= ind2;i++) {
      if (Sequence::sameNuc(_ali1[i],_ali2[i]) > 0) n_id++;
      if (Sequence::isGap(_ali1[i]) || Sequence::isGap(_ali2[i])) n_gap++;
    }
    for (int i = ind3;i <= ind4;i++) {
      if (Sequence::sameNuc(_ali1[i],_ali2[i]) > 0) n_id++;
      if (Sequence::isGap(_ali1[i]) || Sequence::isGap(_ali2[i])) n_gap++;
    }

    cout<<"Score:   "<<setw(9)<<_max<<endl;
    cout<<"Aligned: "<<setw(9)<<n_ali<<"        nucs"<<endl;
    cout<<"Identic: "<<setw(9)<<n_id<<" ("<<setw(3)
	<<(int)(100.*n_id/n_ali + 0.5)<<"%) nucs"<<endl;
    cout<<"Gaps:    "<<setw(9)<<n_gap<<" ("<<setw(3)
	<<(int)(100.*n_gap/n_ali + 0.5)<<"%) nucs"<<endl<<endl;

    int inc1 = 1; if (_s1.reverse()) inc1 = -1;
    int inc2 = 1; if (_s2.reverse()) inc2 = -1;
    int st1 = _s1.start(), st2 = _s2.start();

    // Printing aligned region coordinates
    int first_b1 = st1 + inc1*first_nuc1 - inc1;
    int first_b2 = st1 + inc1*first_nuc2 - inc1;
    int first_b3 = st1 + inc1*first_nuc3 - inc1;
    int first_b4 = st1 + inc1*first_nuc4 - inc1;
    int second_b1 = st2 + inc2*second_nuc1 - inc2;
    int second_b2 = st2 + inc2*second_nuc2 - inc2;
    int second_b3 = st2 + inc2*second_nuc3 - inc2;
    int second_b4 = st2 + inc2*second_nuc4 - inc2;
    int width1 = calcWidth(first_b1,second_b1);
    int width2 = calcWidth(first_b2,second_b2);
    int width3 = calcWidth(first_b3,second_b3);
    int width4 = calcWidth(first_b4,second_b4);
    cout<<"Alignment:"<<endl;
    cout<<" first  seq =>  ";
    if (ind2 >= 0)
      cout<<"["<<setw(width1)<<first_b1<<","<<setw(width2)<<first_b2<<"] ";
    if (ind2 >= 0 && ind3 >= 0) cout<<" "<<UNALIGNED_MESSSAGE<<"  ";
    if (ind3 >= 0)
      cout<<"["<<setw(width3)<<first_b3<<","<<setw(width4)<<first_b4<<"]";
    cout<<endl;
    cout<<" second seq =>  ";
    if (ind2 >= 0)
      cout<<"["<<setw(width1)<<second_b1<<","<<setw(width2)<<second_b2<<"] ";
    if (ind2 >= 0 && ind3 >= 0) cout<<" "<<UNALIGNED_MESSSAGE<<"  ";
    if (ind3 >= 0)
      cout<<"["<<setw(width3)<<second_b3<<","<<setw(width4)<<second_b4<<"]";
    cout<<endl<<endl;

    if (ind2 >= 0 && ind3 >= 0) { // Info about unaligned regions
      int len1 = _rg1 - _lg1 - 1, len2 = _rg2 - _lg2 - 1;
      cout<<UNALIGNED_MESSSAGE<<":"<<endl;
      cout<<" first  seq => "<<setw(9)<<len1<<" nucs";
      if (len1 > 0) cout<<" ["
			<<_s1.start() + inc1*_lg1<<","
			<<_s1.start() + inc1*_rg1 - 2*inc1<<"]";
      cout<<endl;
      cout<<" second seq => "<<setw(9)<<len2<<" nucs";
      if (len2 > 0) cout<<" ["
			<<_s2.start() + inc2*_lg2<<","
			<<_s2.start() + inc2*_rg2 - 2*inc2<<"]";
      cout<<endl;
      
      if (_lg1_alt != _lg1 || _rg1_alt != _rg1 ||
	  _lg2_alt != _lg2 || _rg2_alt != _rg2) {
	len1 = _rg1_alt - _lg1_alt - 1, len2 = _rg2_alt - _lg2_alt - 1;
	cout<<"ALTERNATIVE REGION:"<<endl;
	cout<<" first  seq => "<<setw(9)<<len1<<" nucs";
	if (len1 > 0) cout<<" ["
			  <<_s1.start() + inc1*_lg1_alt<<","
			  <<_s1.start() + inc1*_rg1_alt - 2*inc1<<"]";
	cout<<endl;
	cout<<" second seq => "<<setw(9)<<len2<<" nucs";
	if (len2 > 0) cout<<" ["
			  <<_s2.start() + inc2*_lg2_alt<<","
			  <<_s2.start() + inc2*_rg2_alt - 2*inc2<<"]";
	cout<<endl;
      }
      cout<<endl;
    }

    // Printing homologies around breakpoints
    if (_lg1 > 0 && _lg2 > 0 && _rg1 <= _len1 && _rg2 <= _len2) {
      int left,right;
      int n_hom = findBreakpointHomology(_seq1,_lg1,_rg1,left,right);
      cout<<"Homology at breakpoints: "<<endl;
      cout<<" first  seq => "<<setw(9)<<n_hom<<" nucs";
      if (n_hom > 0)
	cout<<" ["<<_s1.start() + inc1*(_lg1 + left) + inc1<<","
	    <<_s1.start() + inc1*(_lg1 + right)<<"] to"
	    <<" ["<<_s1.start() + inc1*(_rg1 + left)<<","
	    <<_s1.start() + inc1*(_rg1 + right) - inc1<<"]";
      cout<<endl;
      
      n_hom = findBreakpointHomology(_seq2,_lg2,_rg2,left,right);
      cout<<" second seq => "<<setw(9)<<n_hom<<" nucs";
      if (n_hom > 0)
	cout<<" ["<<_s2.start() + inc2*(_lg2 + left) + inc2<<","
	    <<_s2.start() + inc2*(_lg2 + right)<<"] to"
	    <<" ["<<_s2.start() + inc2*(_rg2 + left)<<","
	    <<_s2.start() + inc2*(_rg2 + right) - inc2<<"]";
      cout<<endl;

      n_hom = findOutsideHomology(_seq1,_lg1,_rg1);
      cout<<"Homology outside breakpoints: "<<endl;
      cout<<" first  seq => "<<setw(9)<<n_hom<<" nucs";
      if (n_hom > 0)
	cout<<" ["<<_s1.start() + inc1*(_lg1 - n_hom)<<","
	    <<_s1.start() + inc1*_lg1 - inc1<<"] to"
	    <<" ["<<_s1.start() + inc1*_rg1 - inc1<<","
	    <<_s1.start() + inc1*(_rg1 + n_hom) - 2*inc1<<"]";
      cout<<endl;


      n_hom = findOutsideHomology(_seq2,_lg2,_rg2);
      cout<<" second seq => "<<setw(9)<<n_hom<<" nucs";
      if (n_hom > 0)
	cout<<" ["<<_s2.start() + inc2*(_lg2 - n_hom)<<","
	    <<_s2.start() + inc2*_lg2 - inc2<<"] to"
	    <<" ["<<_s2.start() + inc2*_rg2 - inc2<<","
	    <<_s2.start() + inc2*(_rg2 + n_hom) - 2*inc2<<"]";
      cout<<endl;
      
      n_hom = findInsideHomology(_seq1,_lg1,_rg1);
      cout<<"Homology inside breakpoints: "<<endl;
      cout<<" first  seq => "<<setw(9)<<n_hom<<" nucs";
      if (n_hom > 0)
	cout<<" ["<<_s1.start() + inc1*_lg1<<","
	    <<_s1.start() + inc1*(_lg1 + n_hom) - inc1<<"] to"
	    <<" ["<<_s1.start() + inc1*(_rg1 - n_hom) - inc1<<","
	    <<_s1.start() + inc1*_rg1 - 2*inc1<<"]";
      cout<<endl;


      n_hom = findInsideHomology(_seq2,_lg2,_rg2);
      cout<<" second seq => "<<setw(9)<<n_hom<<" nucs";
      if (n_hom > 0)
	cout<<" ["<<_s2.start() + inc2*_lg2<<","
	    <<_s2.start() + inc2*(_lg2 + n_hom) - inc2<<"] to"
	    <<" ["<<_s2.start() + inc2*(_rg2 - n_hom) - inc2<<","
	    <<_s2.start() + inc2*_rg2 - 2*inc2<<"]";
      cout<<endl<<endl;
      
    }

    // Printing actual alignment
    if (ind2 >= 0) {
      string s1 = _ali1.substr(ind1,ind2 - ind1 + 1);
      string s2 = _ali2.substr(ind1,ind2 - ind1 + 1);
      printSeqAli(s1,s2,first_b1,second_b1,_s1.reverse(),_s2.reverse());
    }
    if (ind3 >= 0) {
      if (ind2 >= 0) cout<<UNALIGNED_MESSSAGE<<endl<<endl;
      string s1 = _ali1.substr(ind3,ind4 - ind3 + 1);
      string s2 = _ali2.substr(ind3,ind4 - ind3 + 1);
      printSeqAli(s1,s2,first_b3,second_b3,_s1.reverse(),_s2.reverse());
    }
  }

  int score() { return _max; }
    
private:
  void setTraceForScore(unsigned char *val,
			unsigned int trace)   { *val |= trace;    }
  void setTraceRevScore(unsigned char *val,
			unsigned int trace)   { *val |= trace<<2; }
  void setTraceForMaximum(unsigned char *val,
			  unsigned int trace) { *val |= trace<<4; }
  void setTraceRevMaximum(unsigned char *val,
			  unsigned int trace) { *val |= trace<<6; }
  unsigned char getTraceForScore(unsigned int val)   { return (val<<30)>>30; }
  unsigned char getTraceRevScore(unsigned int val)   { return (val<<28)>>30; }
  unsigned char getTraceForMaximum(unsigned int val) { return (val<<26)>>30; }
  unsigned char getTraceRevMaximum(unsigned int val) { return (val<<24)>>30; }

  int calcWidth(int v1,int v2)
  {
    int width1 = 0; if (v1 < 0) v1++;
    while (v1 != 0) { width1++; v1 /= 10; }
    int width2 = 0; if (v2 < 0) v2++;
    while (v2 != 0) { width2++; v2 /= 10; }
    if (width1 > width2) return width1;
    return width2;
  }

  void printSeqAli(string &str1,string &str2,
		  int      st1,     int st2,
		  bool  revcom1,bool revcom2)
  {
    static const int WIDTH  = 60;
    static const int MARGIN =  9;

    int inc1 = 1; if (revcom1) inc1 = -1;
    int inc2 = 1; if (revcom2) inc2 = -1;

    string margin = "";
    for (int j = 0;j <= MARGIN;j++) margin += ' ';

    int n = str1.length();
    int ind1 = st1, ind2 = st2;
    for (int i = 0;i < n;i += WIDTH) {
      int start1 = ind1,start2 = ind2;
      string a1 = str1.substr(i,WIDTH);
      string a2 = str2.substr(i,WIDTH);
      int nuc1 = 0,nuc2 = 0;
      string match = "";
      for (int j = 0;j < a1.length();j++) {
	if (Sequence::sameNuc(a1[j],a2[j]) > 0) match += '|';
	else                                    match += ' ';
	if (!Sequence::isGap(a1[j])) { nuc1++; ind1 += inc1; }
	if (!Sequence::isGap(a2[j])) { nuc2++; ind2 += inc2; }
      }

      if (nuc1 > 0) {
	cout<<setw(MARGIN)<<start1<<' '<<a1;
	cout<<' '<<ind1 - inc1<<endl;
      } else cout<<margin<<a1<<endl;

      cout<<margin<<match<<endl;

      if (nuc2 > 0) {
	cout<<setw(MARGIN)<<start2<<' '<<a2;
	cout<<' '<<ind2 - inc2<<endl;
      } else cout<<margin<<a2<<endl;

      cout<<endl;
    }
  }

  int findOutsideHomology(string &seq,int bp1,int bp2)
  {
    int n = seq.length(),d1 = bp1, d2 = n - bp2 + 1;
    int n_check = d1; if (d2 < d1) n_check = d2;
    int start = bp1 - n_check;
    for (int i = 0;i < n_check;i++) {
      int n_same = 0;
      for (int j = i;j < n_check;j++)
	if (Sequence::sameNuc(seq[start + j],seq[bp2 + j - i - 1])) n_same++;
	else break;
      if (n_same == n_check - i) return n_same;
    }

    return 0;
  }

  int findInsideHomology(string &seq,int bp1,int bp2)
  {
    int n = seq.length(),d1 = bp1, d2 = n - bp2 + 1;
    int n_check = (bp2 - bp1)/2, start = bp2 - n_check - 1;
    for (int i = 0;i < n_check;i++) {
      int n_same = 0;
      for (int j = i;j < n_check;j++) {
	if (Sequence::sameNuc(seq[bp1 + j - i],seq[start + j])) n_same++;
	else break;
      }
      if (n_same == n_check - i) return n_same;
    }

    return 0;
  }

  int findBreakpointHomology(string &seq,int bp1,int bp2,int &left,int &right)
  {
    int n = seq.length(),delta = bp2 - bp1 - 1,ret = 0;
    if (delta == 0) return 0;
    int start = bp1 - 1;
    while (start >= 0 && Sequence::sameNuc(seq[start],seq[start + delta])) {
      start--;
      ret++;
    }
    left = start - bp1;

    int end = bp1;
    while (end < n && Sequence::sameNuc(seq[end],seq[end + delta])) {
      end++;
      ret++;
    }
    right = end - bp1 - 1;

    return ret;
  }

  void printMatrix(int *matr)
  {
    long pos = 0;
    for (int i2 = 0;i2 < score_n2;i2++) {
      for (int i1 = 0;i1 < score_n1;i1++)
	cout<<setw(4)<<matr[pos++]<<" ";
      cout<<endl;
    }
    cout<<endl;
  }

  int findUnaligned(int &lg1,int &lg2,int &rg1,int &rg2,bool forward = true)
  {
    int max = 0;
    int ind1 = -1,ind2 = -1;
    unsigned short *f_maxima = f_score, *r_maxima = r_score;
    if (forward) {
      long pos = 0;
      for (int i2 = 0;i2 < score_n2 - 1;i2++) {
	for (int i1 = 0;i1 < score_n1 - 1;i1++) {
	  unsigned short val = f_maxima[pos] + r_maxima[pos + score_n1 + 1];
	  if (val > max) {
	    max = val;
	    ind1 = i1;
	    ind2 = i2;
	  }
	  pos++;
	}
	pos++;
      }
    } else { // Reverse
      long pos = score_size - 1;
      for (int i2 = score_n2 - 1;i2 > 0;i2--) {
	for (int i1 = score_n1 - 1;i1 > 0;i1--) {
	  unsigned short val = r_maxima[pos] + f_maxima[pos - score_n1 - 1];
	  if (val > max) {
	    max = val;
	    ind1 = i1;
	    ind2 = i2;
	  }
	  pos--;
	}
	pos--;
      }
    }

    // Tracing back maxima
    lg1 = ind1; if (!forward) lg1--;
    lg2 = ind2; if (!forward) lg2--;
    long pos = (long)lg2*(long)score_n1 + (long)lg1;
    unsigned char tr = getTraceForMaximum(trace[pos]);
    while (tr != 0) {
      if (tr == VERTICAL) {
	pos -= score_n1;
	lg2--;
      } else if (tr == HORIZONTAL) {
	pos--;
	lg1--;
      } else cerr<<"Internal error (1) in tracking."<<endl;
      tr = getTraceForMaximum(trace[pos]);
    }
    
    rg1 = ind1; if (forward) rg1++;
    rg2 = ind2; if (forward) rg2++;
    pos = (long)rg2*(long)score_n1 + (long)rg1;
    tr = getTraceRevMaximum(trace[pos]);
    while (tr != 0) {
      if (tr == VERTICAL) {
	pos += score_n1;
	rg2++;
      } else if (tr == HORIZONTAL) {
	pos++;
	rg1++;
      } else cerr<<"Internal error (2) in tracking."<<endl;
      tr = getTraceRevMaximum(trace[pos]);
    }
    
    return max;
  }
  
  void findAlignment(Block *bls)
  {
    _ali1 = _ali2 = "";

    int ind1 = 0,ind2 = 0;

    while (bls) {
      int st1 = bls->start1() - 1, st2 = bls->start2() - 1,len = bls->length();
      while (ind1 < st1) { _ali1 += _seq1[ind1++]; _ali2 += Sequence::gap(); }
      while (ind2 < st2) { _ali2 += _seq2[ind2++]; _ali1 += Sequence::gap(); }
      for (int i = 0;i < len;i++) {
	_ali1 += _seq1[ind1++];
	_ali2 += _seq2[ind2++];
      }
      bls = bls->next();
    }
    
    while (ind1 < _len1) { _ali1 += _seq1[ind1++]; _ali2 += Sequence::gap(); }
    while (ind2 < _len2) { _ali2 += _seq2[ind2++]; _ali1 += Sequence::gap(); }
  }

  Block *findBlocks(int left1,int left2,int right1,int right2)
  {
    Block *bls = NULL;

    // Producing alignment blocks for left alignment
    int n_save = _len1; if (_len2 > n_save) n_save = _len2;
    int *p1 = new int[n_save];
    int *p2 = new int[n_save];
    n_save = 0;
    long pos = (long)left2*(long)score_n1 + (long)left1;
    unsigned char tr = getTraceForScore(trace[pos]);
    while (tr != 0) {
      if (tr == DIAGONAL) {
	p1[n_save] = left1--;
	p2[n_save] = left2--;
	n_save++;
	pos -= (score_n1 + 1);
      } else if (tr == HORIZONTAL) {
	left1--;
	pos--;
      } else if (tr == VERTICAL) {
	left2--;
	pos -= score_n1;
      } else cerr<<"Internal error (3) in tracking."<<endl;
      tr = getTraceForScore(trace[pos]);
    }

    int start1,start2;
    if (n_save > 0) {
      int index = n_save - 1,n = 1;
      start1 = p1[index];
      start2 = p2[index];
      index--;
      while (index >= 0) {
	if (p1[index] - p1[index + 1] == 1 &&
	    p2[index] - p2[index + 1] == 1) n++;
	else {
	  Block *b = new Block(start1,start2,n,"left");
	  if (bls == NULL) bls = b;
	  else             bls = bls->next(b);
	  start1 = p1[index];
	  start2 = p2[index];
	  n = 1;
	}
	index--;
      }
      Block *b = new Block(start1,start2,n,"left");
      if (bls == NULL) bls = b;
      else             bls = bls->next(b);
    }
    
    delete[] p1;
    delete[] p2;

    if (bls) while (bls->next()) bls = bls->next(); // Looping to the last

    // Producing alignment blocks for right alignment
    pos = (long)right2*(long)score_n1 + (long)right1;
    int n = 0;
    tr = getTraceRevScore(trace[pos]);
    while (tr != 0) {
      if (tr == DIAGONAL) {
	if (right1 <= _len1 && right2 <= _len2) {
	  if (n == 0) {
	    start1 = right1;
	    start2 = right2;
	    n = 1;
	  } else n++;
	}
	right1++;
	right2++;
	pos += (score_n1 + 1);
      } else if (tr == HORIZONTAL) {
	right1++;
	pos++;
	if (n > 0) {
	  Block *b = new Block(start1,start2,n,"right");
	  if (bls == NULL) bls = b;
	  else             bls = bls->next(b);
	  n = 0;
	}
      } else if (tr == VERTICAL) {
	right2++;
	pos += score_n1;
	if (n > 0) {
	  Block *b = new Block(start1,start2,n,"right");
	  if (bls == NULL) bls = b;
	  else             bls = bls->next(b);
	  n = 0;
	}
      } else cerr<<"Internal error (4) in tracking."<<endl;
      tr = getTraceRevScore(trace[pos]);
    }

    if (n > 0) {
      Block *b = new Block(start1,start2,n,"right");
      if (bls == NULL) bls = b;
      else             bls = bls->next(b);
    }

    while (bls->prev()) bls = bls->prev();
    return bls;
  }

  void calcMaxima()
  {
    // Finding forward maxima
    unsigned short *f_maxima = f_score;
    long pos = score_n1 + 1;
    for (int i2 = 0;i2 < _len2;i2++) {
      for (int i1 = 0;i1 < _len1;i1++) {
	long r_pos = pos - score_n1;
	unsigned char curr_trace = 0;
	if (i2 > 0 && f_maxima[r_pos] > f_maxima[pos]) {
	  f_maxima[pos] = f_maxima[r_pos];
	  curr_trace    = VERTICAL;
	}
	if (i1 > 0 && f_maxima[pos - 1] > f_maxima[pos]) {
	  f_maxima[pos] = f_maxima[pos - 1];
	  curr_trace    = HORIZONTAL;
	}
	setTraceForMaximum(&trace[pos],curr_trace);
	pos++;
      }
      pos += 2;
    }

    // Finding reverse maxima
    unsigned short *r_maxima = r_score;
    pos = score_size - score_n1 - 2;
    for (int i2 = _len2 - 1;i2 >= 0;i2--) {
      for (int i1 = _len1 - 1;i1 >= 0;i1--) {
	long r_pos = pos + score_n1;
	unsigned char curr_trace = 0;
	if (i2 < _len2 - 1 && r_maxima[r_pos] > r_maxima[pos]) {
	  r_maxima[pos] = r_maxima[r_pos];
	  curr_trace    = VERTICAL;
	}
	if (i1 < _len1 - 1 && r_maxima[pos + 1] > r_maxima[pos]) {
	  r_maxima[pos] = r_maxima[pos + 1];
	  curr_trace    = HORIZONTAL;
	}
	setTraceRevMaximum(&trace[pos],curr_trace);
	pos--;
      }
      pos -= 2;
    }
  }

  void calcScores()
  {
    Scorer scr(MATCH,MISMATCH);

    // Filling in the forward direction
    long pos = score_n1 + 1;
    for (int i2 = 0;i2 < _len2;i2++) {
      for (int i1 = 0;i1 < _len1;i1++) {
	long r_pos = pos - score_n1 - 1;
	int  score = f_score[r_pos];
	score += scr.getScore(_seq1[i1],_seq2[i2]);
	int sg1 = f_score[pos - 1];
	unsigned char tr = getTraceForScore(trace[pos - 1]);
	if (tr == HORIZONTAL || tr == VERTICAL) sg1 += GAP_EXTEND;
	else                                    sg1 += GAP_OPEN;
	int sg2 = f_score[r_pos + 1];
	tr = getTraceForScore(trace[r_pos + 1]);
	if (tr == HORIZONTAL || tr == VERTICAL) sg2 += GAP_EXTEND;
	else                                    sg2 += GAP_OPEN;
	unsigned char curr_trace = DIAGONAL;
	if (sg1 >= score) { score = sg1; curr_trace = HORIZONTAL; }
	if (sg2 >= score) { score = sg2; curr_trace = VERTICAL; }
	if (score < 0)    { score = 0;   curr_trace = 0; }
	setTraceForScore(&trace[pos],curr_trace);
	f_score[pos] = (unsigned short)score;
	pos++;
      }
      pos += 2;
    }
    
    // Filling in the reverse direction
    pos = score_size - score_n1 - 2;
    for (int i2 = _len2 - 1;i2 >= 0;i2--) {
      for (int i1 = _len1 - 1;i1 >= 0;i1--) {
	long r_pos = pos + score_n1 + 1;
	int  score = r_score[r_pos];
	score += scr.getScore(_seq1[i1],_seq2[i2]);
	int sg1 = r_score[pos + 1];
	unsigned char tr = getTraceRevScore(trace[pos + 1]);
	if (tr == HORIZONTAL || tr == VERTICAL) sg1 += GAP_EXTEND;
	else                                    sg1 += GAP_OPEN;
	int sg2 = r_score[r_pos - 1];
	tr = getTraceRevScore(trace[r_pos - 1]);
	if (tr == HORIZONTAL || tr == VERTICAL) sg2 += GAP_EXTEND;
	else                                    sg2 += GAP_OPEN;
	unsigned char curr_trace = DIAGONAL;
	if (sg1 >= score) { score = sg1; curr_trace = HORIZONTAL; }
	if (sg2 >= score) { score = sg2; curr_trace = VERTICAL; }
	if (score < 0)    { score = 0;   curr_trace = 0; }
	setTraceRevScore(&trace[pos],curr_trace);
	r_score[pos] = (unsigned short)score;
	pos--;
      }
      pos -= 2;
    }
  }

};

bool parseStartEnd(string &str,int &start,int &end)
{
  int n = str.length(),i = 0;
  string s(""),e("");
  while (i < n && isdigit(str[i])) s += str[i++];
  if (i == n)          return false;
  if (str[i++] != '-') return false;
  while (i < n && isdigit(str[i])) e += str[i++];
  if (i != n)          return false;
  
  int sv = -1,ev = -1;
  if (s.length() > 0) sv = strtol(s.c_str(),NULL,10);
  if (e.length() > 0) ev = strtol(e.c_str(),NULL,10);
  if (sv == 0 || ev == 0)          return false;
  if (sv > 0 && ev > 0 && ev < sv) return false;
  start = sv;
  end   = ev;
  return true;
}

int parseArgs(string line,string *args,int max_args)
{
  int ret = 0, i = 0;
  while (i < line.length()) {
    args[ret] = "";
    while (i < line.length() && line.at(i) == ' ') i++; // Skipping spaces
    while (i < line.length() && line.at(i) != ' ')
      args[ret] += line.at(i++);
    if (args[ret].length() > 0) ret++;
  }

  return ret;
}

int main(int argc,char *argv[])
{
  string exe = argv[0];
  exe = exe.substr(exe.rfind('/') + 1);
  string usage = "Usage:\n\t";
  usage += exe;
  usage += " [-both] [-revcom1] [-revcom2]\n";
  usage += "\t\t[-coor1=start-end] [-coor2=start-end]";
  usage += " file1.fa file2.fa";

  cin.seekg(0,ios::end);
  int cin_length = cin.tellg();
  cin.seekg(0,ios::beg);

  if (argc < 3 && cin_length <= 0) {
    cerr<<usage<<endl;
    return 0;
  }

  const static int MAX_ARGS = 1024;
  string args[MAX_ARGS],line;
  int    n_args = 0;

  if (argc > 1) // If we have input arguments
    for (int i = 1;i < argc;i++) {
      args[n_args] = argv[i];
      line += " ";
      line += args[n_args++];
      if (n_args == MAX_ARGS) {
	cerr<<"Too many input arguments. Skipping!"<<endl;
	break;
      }
    }
  else while (cin_length > 0) { // Taking lines from redirection
      getline(cin,line);
      cin_length -= line.length() + 1;
      if (line.length() > 0) {
	n_args = parseArgs(line,args,MAX_ARGS);
	break;
      }
    }

  string    old_file1(""),    old_file2("");
  Sequence *old_seq1 = NULL, *old_seq2 = NULL;
  while (n_args > 0) {
    string file1(""),file2("");
    int start1 = -1,end1 = -1;
    int start2 = -1,end2 = -1;
    bool revcom1 = false, revcom2 = false, both = false, error = false;
    for (int i = 0;i < n_args;i++) {
      if (args[i].at(0) == '-') { // Options
	string option = args[i];
	string begin = option.substr(0,7);
	if (begin == "-coor1=") {
	  string end = option.substr(7);
	  if (!parseStartEnd(end,start1,end1)) {
	    cerr<<line<<endl;
	    cerr<<"Error in coordinate specification '"<<option<<"'."<<endl;
	    cerr<<usage<<endl;
	    error = true;
	    break;
	  }
	} else if (begin == "-coor2=") {
	  string end = option.substr(7);
	  if (!parseStartEnd(end,start2,end2)) {
	    cerr<<line<<endl;
	    cerr<<"Error in coordinate specification '"<<option<<"'."<<endl;
	    cerr<<usage<<endl;
	    error = true;
	    break;
	  }
	} else if (option == "-revcom1") {
	  revcom1 = true;
	} else if (option == "-revcom2") {
	  revcom2 = true;
	} else if (option == "-both") {
	  both = true;
	} else {
	  cerr<<line<<endl;
	  cerr<<"Unknown option '"<<option<<"'."<<endl;
	  cerr<<usage<<endl;
	  error = true;
	  break;
	}
	continue;
      }
      
      if (file1.length()      <= 0) file1 = args[i];
      else if (file2.length() <= 0) file2 = args[i];
      else {
	cerr<<line<<endl;
	cerr<<"Too many input files '"<<args[i]<<"'."<<endl;
	cerr<<usage<<endl;
	error = true;
	break;
      }
    }

    if (!error) {

      if (file1 != old_file1 || !old_seq1) {
	delete old_seq1;
	old_seq1  = new Sequence(file1);
	old_file1 = file1;
      }
      if (file2 != old_file2 || !old_seq2) {
	delete old_seq2;
	old_seq2  = new Sequence(file2);
	old_file2 = file2;
      }
      
      Sequence *seq1 = old_seq1->substr(start1,end1);
      Sequence *seq2 = old_seq2->substr(start2,end2);
      
      if (revcom1) seq1->revcom();
      if (revcom2) seq2->revcom();

      // Sequence seq1("ACTGGTGTCAACTG"),seq2("ACTGACTG");
      // Sequence seq1("GGACTGGACTG"),   seq2("CACTGGACTG");
      // Sequence seq1("CCCCAGCTTT"),    seq2("AGCCTTT");
      // Sequence seq1("AGCCTTT"),       seq2("AGCCTTT");
      
#ifdef AGE_TIME
      timeval ali_s,ali_e;
      gettimeofday(&ali_s,NULL);
#endif

      if (both) {
	Sequence *seq3 = seq2->clone();
	seq3->revcom();
	AGEaligner aligner1(*seq1,*seq2);
	AGEaligner aligner2(*seq1,*seq3);
	bool res1 = aligner1.align(), res2 = aligner2.align();
	if (!res1 && !res2) cerr<<"No alignment made."<<endl;
	else if (aligner1.score() >= aligner2.score())
	  aligner1.printAlignment();
	else 
	  aligner2.printAlignment();
	delete seq3;
      } else {
	AGEaligner aligner(*seq1,*seq2);
	if (aligner.align())  aligner.printAlignment();
	else cerr<<"No alignment made."<<endl;
      }
      

#ifdef AGE_TIME
    gettimeofday(&ali_e,NULL);
    cout<<"Alignment time is "<<ali_e.tv_sec - ali_s.tv_sec +
      (ali_e.tv_usec - ali_s.tv_usec)/1000000.<<" s"<<endl<<endl;
#endif

      delete seq1;
      delete seq2;
    }

    n_args = 0;
    while (cin_length > 0) {
      getline(cin,line);
      cin_length -= line.length() + 1;
      if (line.length() > 0) {
	n_args = parseArgs(line,args,MAX_ARGS);
	break;
      }
    }
  }    
  delete old_seq1;
  delete old_seq2;


  return 1;
}
